"""Implements vidsz.interfaces.IReader interface for OpenCV Backend.
"""
from typing import Optional, Union, List

import cv2
import numpy as np

from vidsz.interfaces import IReader


class Reader(IReader):
    """Video Reading wrapper around Opencv-Backend
    """
    def __init__(self,
                 source_name: Union[str, int],
                 batch_size: Optional[int] = None,
                 dynamic_batch: Optional[bool] = False,
                 **kwargs) -> None:
        """Initiate Reader object

        Args:
            source_name (Union[str, int]): Name/URL/Path to video source

            batch_size (Optional[int]): number of frames to return (as one batch) for one read.
            Defaults to None will return images individually without batch axis.

            dynamic_batch (Optional[bool]): if set to True then last batch of frames may have
            less than batch_size frames (depending on how many frames were left for last batch).
            If set to False, last batch may have some frames made up of zeros to match batch_size.
            Defaults to False.
        """
        # initiate props
        self._initiate_props()

        # sanity conversion
        self._name = str(source_name)

        # set batch
        self._batch_size = batch_size
        self._dynamic_batch = dynamic_batch

        # open video stream
        self._video_stream = cv2.VideoCapture(
            int(self._name) if self._name.isdigit() else self._name)

        # update info with current video stream
        self._update_info()

    def _initiate_props(self) -> None:
        """Initiate all class properties to default values
        """
        self._name = None
        self._is_open = True
        self._video_stream = None
        self._width = None
        self._height = None
        self._fps = None
        self._backend = 'opencv'
        self._info = None
        self._frame_count = 0
        self._seconds = 0
        self._minutes = 0
        self._batch_size = None
        self._dynamic_batch = False

    def _update_info(self) -> None:
        """Update info property according to currently open video stream

        Raises:
            Exception: VideoSourceNotOpen raised when no video stream is opened
        """
        # check if source is open
        if not self.is_open():
            raise Exception((f"Failed to read from {self.name}. " +
                             "Please check the filename/source-info again."))

        # update relevant props
        self._width = int(self._video_stream.get(cv2.CAP_PROP_FRAME_WIDTH))
        self._height = int(self._video_stream.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self._fps = self._video_stream.get(cv2.CAP_PROP_FPS)
        self._is_open = bool(self._video_stream.isOpened())

        # update info
        self._info = {
            "name": self._name,
            "width": self._width,
            "height": self._height,
            "fps": self._fps,
            "backend": self._backend
        }

    @property
    def name(self) -> str:
        """Name of Video Source

        Returns:
            str: name of video source
        """
        return self._name

    @property
    def width(self) -> int:
        """Width of Video

        Returns:
            int: width of video frame
        """
        return self._width

    @property
    def height(self) -> int:
        """Height of Video

        Returns:
            int: height of video frame
        """
        return self._height

    @property
    def fps(self) -> float:
        """FPS of Video

        Returns:
            float: fps of video
        """
        return self._fps

    @property
    def backend(self) -> str:
        """Name of the Backend being used

        Returns:
            str: current backend name
        """
        return self._backend

    @property
    def info(self) -> dict:
        """Video information

        Returns:
            dict: info of width, height, fps and backend.
        """
        return self._info

    @property
    def frame_count(self) -> int:
        """Total frames read

        Returns:
            int: read frames' count
        """
        return self._frame_count

    @property
    def seconds(self) -> float:
        """Total seconds read

        Returns:
            float: read frames' in seconds
        """
        return (self._frame_count / self._fps) if self._fps else 0

    @property
    def minutes(self) -> float:
        """Total minutes read

        Returns:
            float: read frames' in minutes
        """
        return self.seconds / 60.0

    def is_open(self) -> bool:
        """Checks if video is still open and last read frame was valid

        Returns:
            bool: True if video is open and last frame was not None, false otherwise.
        """
        return self._video_stream.isOpened() and self._is_open

    def read_frame(self) -> Union[np.ndarray, None]:
        """Returns next frame from the video if available

        Returns:
            Union[np.ndarry, None]: next frame if available, None otherwise.
        """
        flag, frame = self._video_stream.read()
        self._frame_count += 0 if frame is None else 1
        self._is_open = flag
        return frame

    def read_batch(self) -> Union[np.ndarray, None]:
        """Returns next batch of frames from the video if available

        Returns:
            Union[np.ndarry, None]: next batch if available, None otherwise.
        """
        if not self.is_open():
            return None

        # pre-allocate batch
        batch = np.zeros((self._batch_size, self.height, self.width, 3), dtype="uint8")

        # fill batch
        for i in range(self._batch_size):
            # read frame
            frame = self.read_frame()

            # stop process, no frames left
            if frame is None:
                # decrm index because this frame was empty
                i -= 1
                break

            # add to batch
            batch[i] = frame

        return batch[:i + 1] if self._dynamic_batch else batch

    def read(self) -> Union[np.ndarray, None]:
        """Returns next frame or batch of frames from the video if available

        Returns:
            Union[np.ndarry, None]: next frame or batch of frames if available, None otherwise.
        """
        if self._batch_size is None:
            return self.read_frame()
        return self.read_batch()

    def release(self) -> None:
        """Release Resources
        """
        if self._video_stream is not None:
            self._video_stream.release()

    def __del__(self) -> None:
        """Release Resources
        """
        self.release()
        self._video_stream = None

    def __next__(self) -> np.ndarray:
        """Returns next frame from the video

        Raises:
            StopIteration: No more frames to read

        Returns:
            np.ndarray: frame read from video
        """
        frame = self.read()
        if frame is None:
            raise StopIteration()
        return frame

    def __iter__(self) -> "Reader":
        """Returns iterable object for reading frames

        Returns:
            Iterable[Reader]: iterable object for reading frames
        """
        return self

    def __repr__(self) -> str:
        """Video's Info

        Returns:
            str: info
        """
        return str(self._info)

    def __str__(self) -> str:
        """Video's Info

        Returns:
            str: Info
        """
        return str(self._info)

    def __enter__(self) -> "Reader":
        """Returns Conext for "with" block usage

        Returns:
            Reader: Video Reader object
        """
        return self

    def __exit__(self, exc_type: None, exc_value: None,
                 traceback: None) -> None:
        """Release resources before exiting the "with" block

        Args:
            exc_type (NoneType): Exception type if any
            exc_value (NoneType): Exception value if any
            traceback (NoneType): Traceback of Exception
        """
        self.release()



"""Implements vidsz.interfaces.IWriter interface for OpenCV Backend.
"""
import os
from typing import Optional, List, Union

import cv2
import numpy as np

from vidsz.interfaces import IWriter
from vidsz.interfaces import IReader


class Writer(IWriter):
    """Video Writing wrapper around Opencv-Backend
    """

    # opencv fourcc mappings
    _EXT_TO_FOURCC = {".avi": "DIVX", ".mkv": "X264", ".mp4": "mp4v"}

    def __init__(self,
                 reader: Optional[IReader] = None,
                 name: Optional[str] = None,
                 width: Optional[int] = None,
                 height: Optional[int] = None,
                 fps: Optional[int] = None,
                 ext: Optional[str] = None) -> None:
        """Initiate Writer object

        Args:
            reader (Optional[IReader], optional): Source for setting
            output video's configs. Defaults to None.

            name (Optional[str], optional): name of output video,
            overwrites reader's. Defaults to None.

            width (Optional[int], optional): width of output video,
            overwrites reader's. Defaults to None.

            height (Optional[int], optional): height of output video.
            overwrites reader's. Defaults to None.

            fps (Optional[int], optional): fps of output video,
            overwrites reader's. Defaults to None.

            ext (Optional[str], optional): extension of output video,
            overwrites reader's. Defaults to None.
        """

        # initiate props
        self._initiate_props()

        # update props if provided
        self._update_props(reader, name, width, height, fps, ext)

        # initiate writer
        self._video_writer = cv2.VideoWriter(self._name, self._fourcc(),
                                             self._fps,
                                             (self._width, self._height))

        # check if open
        if not self.is_open():
            raise AssertionError(
                "Failed to Create Writer for the given settings.")

        # update info
        self._update_info()

    def _initiate_props(self) -> None:
        """Initiate all class properties to default values
        """
        self._name = None
        self._video_writer = None
        self._width = None
        self._height = None
        self._fps = None
        self._backend = "opencv"
        self._ext = None
        self._info = None
        self._frame_count = 0
        self._seconds = 0
        self._minutes = 0

    def _update_props(self,
                      reader: Optional[IReader] = None,
                      name: Optional[str] = None,
                      width: Optional[int] = None,
                      height: Optional[int] = None,
                      fps: Optional[int] = None,
                      ext: Optional[str] = None) -> None:
        """Update all relevant class properties

        Args:
            reader (Optional[IReader], optional): Source for setting
            output video's configs. Defaults to None.

            name (Optional[str], optional): name of output video,
            overwrites reader's. Defaults to None.

            width (Optional[int], optional): width of output video,
            overwrites reader's. Defaults to None.

            height (Optional[int], optional): height of output video.
            overwrites reader's. Defaults to None.

            fps (Optional[int], optional): fps of output video,
            overwrites reader's. Defaults to None.

            ext (Optional[str], optional): extension of output video,
            overwrites reader's. Defaults to None.

        Raises:
            Exception: Raised if neither reader nor name is given.
        """

        # set default if given
        if reader is not None:
            reader_name, reader_ext = os.path.splitext(reader.name)
            self._name = f"{reader_name}_out{reader_ext}"
            self._width = reader.width
            self._height = reader.height
            self._fps = reader.fps

        # override props if given
        if name is not None:
            self._name = name
        if width is not None:
            self._width = width
        if height is not None:
            self._height = height
        if fps is not None:
            self._fps = fps

        # check valid name
        if self._name is None:
            raise AssertionError("Must provide either Reader or name arg.")

        # set ext for CV2 writer creations
        self._ext = os.path.splitext(self._name)[1] if ext is None else ext

        # add . for consistency
        if '.' not in self._ext:
            self._ext = '.' + self._ext

    def _update_info(self) -> None:
        """Update info property according to class props
        """
        # update info
        self._info = {
            "name": self._name,
            "width": self._width,
            "height": self._height,
            "fps": self._fps,
            "backend": self._backend,
            "ext": self._ext
        }

    def _fourcc(self) -> cv2.VideoWriter_fourcc:
        """Returns CV2 VideoWriter_fourcc for writer's ext

        Raises:
            NotImplementedError: raise if unsupported ext is used.

        Returns:
            cv2.VideoWriter_fourcc: fourcc of used ext
        """
        if self._ext not in self._EXT_TO_FOURCC:
            raise NotImplementedError(f"'{self._ext}'is not supported.")
        return cv2.VideoWriter_fourcc(*self._EXT_TO_FOURCC[self._ext])

    @property
    def name(self) -> str:
        """Name of Output Video

        Returns:
            str: name of output video
        """
        return self._name

    @property
    def width(self) -> int:
        """Width of Output Video

        Returns:
            int: width of video frame
        """
        return self._width

    @property
    def height(self) -> int:
        """Height of Output Video

        Returns:
            int: height of video frame
        """
        return self._height

    @property
    def fps(self) -> float:
        """FPS of Output Video

        Returns:
            float: fps of video
        """
        return self._fps

    @property
    def backend(self) -> str:
        """Name of the Backend being used

        Returns:
            str: current backend name
        """
        return self._backend

    @property
    def ext(self) -> str:
        """Extension of Output Video

        Returns:
            str: ext of video
        """
        return self._ext

    @property
    def info(self) -> dict:
        """Video information

        Returns:
            dict: info of width, height, fps, backend and ext
        """
        return self._info

    @property
    def frame_count(self) -> int:
        """Total frames written

        Returns:
            int: written frames' count
        """
        return self._frame_count

    @property
    def seconds(self) -> float:
        """Total seconds written

        Returns:
            float: written frames' in seconds
        """
        return (self._frame_count / self._fps) if self._fps else 0

    @property
    def minutes(self) -> float:
        """Total minutes written

        Returns:
            float: written frames' in minutes
        """
        return self.seconds / 60.0

    def is_open(self) -> bool:
        """Checks if writer is still open

        Returns:
            bool: True if writer is open, False otherwise
        """
        return self._video_writer.isOpened()

    def write(self, frame: np.ndarray) -> None:
        """Write frame to output video

        Args:
            frame (np.ndarray): frame to write

        Raises:
            Exception: raised when method is called on a non-open writer.
        """
        # check if writer is open
        if not self.is_open():
            raise Exception(
                "[Vidsz-Error] Attempted writing with a non-open Writer.")

        self._video_writer.write(frame)
        self._frame_count += 1

    def write_all(self, frames: Union[List[np.ndarray], IReader]) -> None:
        """Write all frames to output video

        Args:
            frames (Union[List[np.ndarray], IReader]): Iterable object that contains frames.
        """
        for frame in frames:
            self.write(frame)

    def release(self) -> None:
        """Release Resources
        """
        if self._video_writer is not None:
            self._video_writer.release()

    def __del__(self) -> None:
        """Release Resources
        """
        self.release()
        self._video_writer = None

    def __repr__(self) -> str:
        """Writer's Info
        Returns:
            str: info
        """
        return str(self.info)

    def __str__(self) -> str:
        """Writer's Info
        Returns:
            str: info
        """
        return str(self.info)

    def __enter__(self) -> "Writer":
        """Returns Conext for "with" block usage
        Returns:
            Writer: Video Reader object
        """
        return self

    def __exit__(self, exc_type: None, exc_value: None,
                 traceback: None) -> None:
        """Release resources before exiting the "with" block
        Args:
            exc_type (NoneType): Exception type if any
            exc_value (NoneType): Exception value if any
            traceback (NoneType): Traceback of Exception
        """
        self.release()



"""Defines interface for video writer
"""
import abc
from typing import List, Union

import numpy as np

from .reader import IReader


class IWriter(metaclass=abc.ABCMeta):
    """Video Writing Interface which will be implemented
    for every supported backend.
    """
    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of Output Video

        Returns:
            str: name of output video
        """
        ...

    @property
    @abc.abstractmethod
    def width(self) -> int:
        """Width of Output Video

        Returns:
            int: width of video frame
        """
        ...

    @property
    @abc.abstractmethod
    def height(self) -> int:
        """Height of Output Video

        Returns:
            int: height of video frame
        """
        ...

    @property
    @abc.abstractmethod
    def fps(self) -> float:
        """FPS of Output Video

        Returns:
            float: fps of video
        """
        ...

    @property
    @abc.abstractmethod
    def backend(self) -> str:
        """Name of the Backend being used

        Returns:
            str: current backend name
        """
        ...

    @property
    @abc.abstractmethod
    def ext(self) -> str:
        """Extension of Output Video

        Returns:
            str: ext of video
        """
        ...

    @property
    @abc.abstractmethod
    def info(self) -> dict:
        """Video information

        Returns:
            dict: info of width, height, fps, backend and ext
        """
        ...

    @property
    @abc.abstractmethod
    def frame_count(self) -> int:
        """Total frames written

        Returns:
            int: written frames' count
        """
        ...

    @property
    @abc.abstractmethod
    def seconds(self) -> float:
        """Total seconds written

        Returns:
            float: written frames' in seconds
        """
        ...

    @property
    @abc.abstractmethod
    def minutes(self) -> float:
        """Total minutes written

        Returns:
            float: written frames' in minutes
        """
        ...

    @abc.abstractmethod
    def is_open(self) -> bool:
        """Checks if writer is still open

        Returns:
            bool: True if writer is open, False otherwise
        """
        ...

    @abc.abstractmethod
    def write(self, frame: np.ndarray) -> None:
        """Write frame to output video

        Args:
            frame (np.ndarray): frame to write
        """
        ...

    @abc.abstractmethod
    def write_all(self, frames: Union[List[np.ndarray], IReader]) -> None:
        """Write all frames to output video

        Args:
            frames (Union[List[np.ndarray], IReader]): Iterable object that contains frames.
        """
        ...

    @abc.abstractmethod
    def release(self) -> None:
        """Release Resources
        """
        ...

    @abc.abstractmethod
    def __del__(self) -> None:
        """Release Resources
        """
        ...

    @abc.abstractmethod
    def __repr__(self) -> str:
        """Writer's Info
        Returns:
            str: info
        """
        ...

    @abc.abstractmethod
    def __str__(self) -> str:
        """Writer's Info
        Returns:
            str: info
        """
        ...

    @abc.abstractmethod
    def __enter__(self) -> "IWriter":
        """Returns Conext for "with" block usage
        Returns:
            IWriter: Video Reader object
        """
        ...

    @abc.abstractmethod
    def __exit__(self, exc_type: None, exc_value: None,
                 traceback: None) -> None:
        """Release resources before exiting the "with" block
        Args:
            exc_type (NoneType): Exception type if any
            exc_value (NoneType): Exception value if any
            traceback (NoneType): Traceback of Exception
        """
        ...



"""Common Wrapper/Interface around various video reading/writing tools to make video reading stable,
consistent and super easy around different systems and OS.

Backends
    - OpenCV (in-development)

# How to Read Video
```python

    from vidsz import Reader

    # open reader
    reader = Reader("static/countdown.mp4")

    # print info: width, height, fps etc.
    print(reader)

    # access specific things
    print(reader.width, reader.height, reader.fps)

    # access number-of-frames/seconds/minutes that have been read
    print(reader.frame_count, reader.seconds, reader.minutes)

    # read frames with while loop
    while reader.is_open():
        # returns ndarry-frame or None if nothing left to read
        frame = reader.read()

    # read frame with for loop
    for frame in reader:
        # use frame however you like
        pass

    # release
    reader.release()


    # or read with a with block
    with Reader("static/countdown.mp4") as reader:
        frame = reader.read()

```

# How to Write Video

```python

    from vidsz import Reader, Writer

    video_fname = "static/countdown.mp4"

    # open reader
    reader = Reader(video_fname)

    # start writer with the Reader object
    # by default it'll append _out in the name of the output video
    writer = Writer(reader)

    # start writer with your settings;
    # you can also give any combinations of
    # following settings with Reader object to
    # overwrite default settings
    writer = Writer("out.mp4", width=680, height=340, fps=30, ext="avi")

    # print writer info
    print(writer)

    # write single frame
    frame = reader.read()
    writer.write(frame)

    # write list of frames
    # or directly write everything from reader object
    writer.write_all(reader)

    # close off
    writer.release()

    # using "with" block, write "static/countdown_out.mp4" (clone of input)
    with Reader(video_fname) as reader:
        with Writer(reader) as writer:
            writer.write_all(reader)
```
"""
__version__ = "0.2.0"



"""Defines interface for video reader
"""
import abc
from typing import Union, List

import numpy as np


class IReader(metaclass=abc.ABCMeta):
    """Video Reader Interface which will be implemented
    for every supported backend.
    """
    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of Video Source

        Returns:
            str: name of video source
        """
        ...

    @property
    @abc.abstractmethod
    def width(self) -> int:
        """Width of Video

        Returns:
            int: width of video frame
        """
        ...

    @property
    @abc.abstractmethod
    def height(self) -> int:
        """Height of Video

        Returns:
            int: height of video frame
        """
        ...

    @property
    @abc.abstractmethod
    def fps(self) -> float:
        """FPS of Video

        Returns:
            float: fps of video
        """
        ...

    @property
    @abc.abstractmethod
    def backend(self) -> str:
        """Name of the Backend being used

        Returns:
            str: current backend name
        """
        ...

    @property
    @abc.abstractmethod
    def info(self) -> dict:
        """Video information

        Returns:
            dict: info of width, height, fps and backend.
        """
        ...

    @property
    @abc.abstractmethod
    def frame_count(self) -> int:
        """Total frames read

        Returns:
            int: read frames' count
        """
        ...

    @property
    @abc.abstractmethod
    def seconds(self) -> float:
        """Total seconds read

        Returns:
            float: read frames' in seconds
        """
        ...

    @property
    @abc.abstractmethod
    def minutes(self) -> float:
        """Total minutes read

        Returns:
            float: read frames' in minutes
        """
        ...

    @abc.abstractmethod
    def is_open(self) -> bool:
        """Checks if video is still open and last read frame was valid

        Returns:
            bool: True if video is open and last frame was not None, false otherwise.
        """
        ...

    @abc.abstractmethod
    def read(self) -> Union[np.ndarray, None]:
        """Returns next frame from the video if available

        Returns:
            Union[np.ndarry, None]: next frame if available, None otherwise.
        """
        ...

    @abc.abstractmethod
    def release(self) -> None:
        """Release Resources
        """
        ...

    @abc.abstractmethod
    def __del__(self) -> None:
        """Release Resources
        """
        ...

    @abc.abstractmethod
    def __next__(self) -> np.ndarray:
        """Returns next frame from the video

        Raises:
            StopIteration: No more frames to read

        Returns:
            np.ndarray: frame read from video
        """
        ...

    @abc.abstractmethod
    def __iter__(self) -> "IReader":
        """Returns iterable object for reading frames

        Returns:
            Iterable[IReader]: iterable object for reading frames
        """
        ...

    @abc.abstractmethod
    def __repr__(self) -> str:
        """Video's Info

        Returns:
            str: info
        """
        ...

    @abc.abstractmethod
    def __str__(self) -> str:
        """Video's Info

        Returns:
            str: Info
        """
        ...

    @abc.abstractmethod
    def __enter__(self) -> "IReader":
        """Returns Conext for "with" block usage

        Returns:
            IReader: Video Reader object
        """
        ...

    @abc.abstractmethod
    def __exit__(self, exc_type: None, exc_value: None,
                 traceback: None) -> None:
        """Release resources before exiting the "with" block

        Args:
            exc_type (NoneType): Exception type if any
            exc_value (NoneType): Exception value if any
            traceback (NoneType): Traceback of Exception
        """
        ...



import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(name="vidsz-deployment-test",
      version="0.1.0",
      description=
      "Common Wrapper/Interface around various video reading/writing tools",
      long_description=README,
      long_description_content_type="text/markdown",
      url="https://github.com/BlueMirrors/vidsz",
      author="BlueMirrors",
      author_email="contact.bluemirrors@gmail.com",
      license="Apache Software License v2.0",
      classifiers=[
          "License :: OSI Approved :: Apache Software License",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
      ],
      packages=find_packages(exclude=("test", )),
      install_requires=["opencv-python"])
